/**
   @file     
   @ingroup alloc
   @brief   libbpm waveform memory allocation routines
   
   This header contains the definitions for the memory allocation routines
   to handle waveforms in libbpm.
*/

/** @addtogroup alloc */
/** @{ */

#ifndef BPMALLOC_H__
#define BPMALLOC_H__

/* -----------------------------------------------------------------------------
// includes
// -------------------------------------------------------------------------- */
#include <stdlib.h>
#include <bpm/bpm_defs.h>

/* -----------------------------------------------------------------------------
// macro definitions
// -------------------------------------------------------------------------- */

/* -----------------------------------------------------------------------------
// typedefs, enums and other declarations
// -------------------------------------------------------------------------- */
#ifdef __cplusplus
extern "C" {
#endif


/* -----------------------------------------------------------------------------
// function prototypes and declarations
// -------------------------------------------------------------------------- */

  /**
     Allocates memory for a complex waveform of doubles. A pointer is returned
     to the reserved memory. Use free_complex_wave_double( w, ns ) to free up
     the memory.

     @param ns  Number of samples
     @return a pointer to the reserved memory
  */
  EXTERN double** alloc_complex_wave_double( int ns );


  /**
     Frees up the memory reserved by alloc_complex_wave_double( ns ).

     @param w   A pointer to a complex waveform of doubles
     @param ns  Number of samples
     @return void
  */
  EXTERN void     free_complex_wave_double( double** w, int ns );


  /**
     Allocates memory for a simple waveform of doubles. A pointer is returned
     to the reserved memory. Use free_simple_wave_double( w ) to free up
     the memory.

     @param ns  Number of samples
     @return a pointer to the reserved memory
  */
  EXTERN double*  alloc_simple_wave_double( int ns ); // alloc simple wave of doubles


  /**
     Frees up the memory reserved by alloc_simple_wave_double( ns ).

     @param w   A pointer to a complex waveform of doubles
     @return void
  */
  EXTERN void     free_simple_wave_double( double* w );


  /**
     Allocates memory for a simple waveform of integers. A pointer is returned
     to the reserved memory. Use free_simple_wave_int( w ) to free up
     the memory.

     @param ns  Number of samples
     @return a pointer to the reserved memory
  */
  EXTERN int*     alloc_simple_wave_int( int ns );  // alloc simple wave of integers


  /**
     Frees up the memory reserved by alloc_simple_wave_int( ns ).

     @param w   A pointer to a complex waveform of integers
     @return void
  */
  EXTERN void     free_simple_wave_int( int *w );

#ifdef __cplusplus
}
#endif

#endif /* #ifndef BPMUNITS_H_LOADED */
/*@}*/
/* ================================ end of file ============================= */
