/**
   @file    
   @ingroup interface
   @brief    Front end interface structure definitions and handlers.

   This header contains the front-end interface structures and handlers
   for libbpm. They define a set of user friendly structures like
   bpmconf_t, bpmcalib_t, beamconf_t etc... to work with the bpm data.
*/

/** @addtogroup interface */
/** @{ */

#ifndef BPMINTERFACE_H__
#define BPMINTERFACE_H__

/* -----------------------------------------------------------------------------
// includes
// -------------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <bpm/bpm_defs.h>

/* -----------------------------------------------------------------------------
// macro definitions
// -------------------------------------------------------------------------- */
#define NMAX_MODES  10

/* -----------------------------------------------------------------------------
// typedefs, enums and other declarations
// -------------------------------------------------------------------------- */
#ifdef __cplusplus
extern "C" {
#endif

  /**
     BPM cavity ( of better signal ) type
  */
  enum bpmtype_t  { 
    diode,    /**< diodified bpm signal or trigger pulse */
    monopole, /**< reference cavity signal ( monopole ) */
    dipole    /**< position sentivive cavity signal ( dipole ) */
  };
  
  /**
     BPM polarisation plane, basically a difficult way to say x or y ;)
  */
  enum bpmpol_t   { 
    horiz, /**< Horizontal plane, or x in most cases*/
    vert   /**< Vertical plane, or y in most cases */
  };
  
  /**
     BPM electronics phase lock type
  */
  enum bpmphase_t { 
    randomised,  /**< unlocked phase */
    locked       /**< locked phase */
  };
  

  typedef struct bpmconf   bpmconf_t;   /**< type definition for BPM configuration */
  typedef struct bpmsignal bpmsignal_t; /**< type definition for BPM signals */
  typedef struct bpmcalib  bpmcalib_t;  /**< type definition for calibrations */
  typedef struct bpmproc   bpmproc_t;   /**< type definition for processed BPM signals */
  typedef struct beamconf  beamconf_t;  /**< type definition for beam configurations */
  typedef struct bpmmode   bpmmode_t;   
  typedef struct rfmodel   rfmodel_t;
  

  /**
     Structure containing the BPM configuration
  */
  struct bpmconf {
    char     name[20];              /**< a BPM should have a name */
    
    enum bpmtype_t  cav_type;         /**< BPM type */
    enum bpmpol_t   cav_polarisation; /**< BPM polarisation */
    enum bpmphase_t cav_phasetype;    /**< BPM phase type */

    rfmodel_t *cav_model;
    double     cav_length;          /**< length of the cavity */
    
    double     cav_freq;            /**< cavity freq (MHz) */
    double     cav_decaytime;       /**< cavity decay time (microsec) */
    double     cav_phase;           /**< phase advance wrt. reference (fixed or random) */
    double     cav_iqrotation;      /**< cavity IQ rotation */
    double     cav_chargesens;      /**< charge sensitivity (volt/nC) */
    double     cav_possens;         /**< position sensitivity at 1.6nC charge (volt/micron) */
    double     cav_tiltsens;        /**< tilt sensitivity at 1.6nC charge (volt/micron) */
    
    double     rf_LOfreq;           /**< LO frequency to mix down with (in MHz) */
    
    double     digi_trigtimeoffset; /**< time (usec) to offset bunch arrival times by */
    double     digi_freq;           /**< digitization frequency (MHz) */
    int        digi_nbits;          /**< number of bits in ADC for digitisation */
    int        digi_nsamples;       /**< number of samples in ADC digitisation */
    double     digi_ampnoise;       /**< amplitude noise in ADC channels (pedestal width) */
    int        digi_voltageoffset;  /**< voltage offset (pedestal position) in counts */
    double     digi_phasenoise;     /**< phase noise */
    
    
    double     geom_pos[3];         /**< position of the BPM in the beamline  */
    double     geom_tilt[3];        /**< tilt of the BPM (0: xrot, 1: yrot, 2: zrot) */
    
    int        ref_idx;             /**< reference cavity index for this BPM */
    int        diode_idx;           /**< reference diode index for this BPM */
  };

  
  /**
     A structure holding the BPM signal
  */
  struct bpmsignal {
    int             *wf;      /**< BPM signal  */
    int              ns;      /**< Number of samples for the waveform (just in case) */
  };


  /**
     A structure containing the calibration information
  */
  struct bpmcalib {
    double   freq;            /**< frequency of downmixed waveform (MHz) */
    double   tdecay;          /**< decay time (usec) */
    double   ddcfiltBW;       /**< ddc filter bandwidth in MHz */
    double   ddcepsFilt;      /**< ddc epsilon filter */
    double   t0Offset;        /**< always have offset from t0 for sampling !!! */
    double   IQphase;         /**< processed IQ phase */
    double   posscale;        /**< processed position scale */
    double   slopescale;      /**< processed slope scale */
  };

  /**
     A structure containing the processed waveform information
  */
  struct bpmproc {
    
    double   ampnoise;        /**< calculated (processed) amplitude noise */
    double   voltageoffset;   /**< calculated voltage offset */
    
    double   t0;              /**< trigger t0 signal */
    
    double **ddcwf;           /**< The digially down converted waveform */
    double **fftwf;           /**< The fourier transform of the waveform */
    
    int      fft_success;     /**< do we have proper fft info ? */
    double   fft_freq;        /**< frequency obtained from fft (MHz) */
    double   fft_tdecay;      /**< decay time obtained from fft (usec) */
    
    int      ddc_success;     /**< do we have proper ddc info ? */
    double   ddc_Q;           /**< ddc Q value */
    double   ddc_I;           /**< ddc I value */
    double   ddc_amp;         /**< downconverted amplitude */
    double   ddc_phase;       /**< downconverted phase */
    double   ddc_tdecay;      /**< downconverted decay time of waveform */
    double   ddc_pos;         /**< calculated position from ddc */
    double   ddc_slope;       /**< calculated slope from ddc */
    
    int      fit_success;     /**< do we have proper fit info ? */
    double   fit_Q;           /**< fit Q value */
    double   fit_I;           /**< fit I value */
    double   fit_amp;         /**< fitted amplitude */
    double   fit_phase;       /**< fitted phase  */
    double   fit_freq;        /**< fitted frequency (MHz) */
    double   fit_tdecay;      /**< fitted decay time of waveform (usec) */
    double   fit_pos;         /**< calculated position from fit */
    double   fit_slope;       /**< calculated slope from fit */
  };
  
  
  /** 
      This structure contains the beam information at a certain point of the orbit.
  */	
  struct beamconf {
    double   energy;          /**< average beam energy (in GeV) */
    double   sig_energy;      /**< energy spread (sigma) */
    double   charge;          /**< bunch charge */
    double   sig_charge;      /**< charge spread (sigma) */

    double   bunchlength;     /**< the bunch length */
    
    double   arrival_time;    /**< arrival time of bunch */
    
    double   beampos[2];      /**< the beam position x,y at the bpm coo */
    double   beamslope[2];    /**< the beam slope x',y' at the bpm coo */
    double   bunchtilt[2];     /**< the bunch tilt x',y' at the bpm coo */
    
    double   bpmhit[3];       /**< where the beam hits the BPM in the BPM local co */
    double   bpmslope[2];     /**< slope of beam through the BPM in BPM local co */   
    double   bpmtilt[2];      /**< bunch tilt in the BPM local co */
  };
  
  

 /**
     This structure defines a BPM resonant mode which is defined by 
     it's resonant frequency, Q factor and sensitivities to the
     beam charge, slope and bunch tilt. */
  struct bpmmode {
    char          name[20];     /**< The name for the BPM mode, e.g "dipolex" */
    double        frequency;    /**< The resonant frequency of the mode */
    double        Q;            /**< The Q factor for the mode */
    int           order;        /**< The mode order, 0:monopole, 1:dipole, 2:quadrupole... */
    enum bpmpol_t polarisation; /**< The mode polarisation: horiz, vert */
    double        sensitivity;  /**< The sensitivity of the mode, units depend on order */
  };

  /**
     This structure contains the complete RF model for a BPM, which is
     essentially a collection of it's resonant modes and sensitivities */
  struct rfmodel {
    char       name[20];          /**< A name for the cavity's RF model */
    int        nmodes;            /**< The number of BPM modes in the model */
    bpmmode_t *mode[NMAX_MODES];  /**< A list of pointers to the array of modes */
  };




/* -----------------------------------------------------------------------------
   function prototypes and declarations
   -------------------------------------------------------------------------- */
  EXTERN int bpm_verbose; /**< be a bit verbose in libbpm */

  /**
     Load a set of bpm configurations from file fname. Memory is allocated using calloc and
     so is the responsibility of the user to delete after use.

     The configuration file lists the fields and their initial values. The first non-comment 
     line is the header for the configuration. Hashed lines indicate comments.

     Example of a bpmconf file:
     
     # Header - libbpm version, number of BPMs
     0.1    21
     
     # Here are the BPM definitions themselves. Add whichever you want though the 
     # ** fields are required.
     # Everything else will be set to -DBL_MAX
     bpm_x9                                      # BPM name. Currently not used.
     {
        bpm_idx               0                 # The index in the created array  **
	
	cav_type              dipole            
        cav_polarisation      horiz             
	cav_phasetype         locked
	cav_freq              2626              # Cavity frequency (in MHz)
	cav_decaytime         3                 # Decay time (microsec)
	cav_phase             0
	cav_iqrotation        0
	cav_chargesens        10
	cav_possens           10
	cav_tiltsens          10

	rf_LOfreq             2550

	digi_trigtimeoffset   50
	digi_freq             100
	digi_nbits            14                # Number of bits in the ADC **
	digi_nsamples         256               # Number of samples in the ADC **
	digi_ampnoise         5
	digi_voltageoffset    8192
	digi_phasenoise       3

	geom_pos              0 0 40
	geom_tilt             0 0 0

	ref_idx               10               # Refernce index **
	diode_idx             10               # Diode index **
    }

    # etc...
    
    @param fname the filename of the configuration file to load
    @param conf the pointer to the newly created set of configurations
    @param num_conf the number of conifgurations loaded

    @return BPM_SUCCESS upon success, BPM_FAILURE upon failure
  */
  EXTERN int load_bpmconf( const char *fname, bpmconf_t **conf, int *num_conf );

  
  /**
     Load in the header information from a configuration file. The header must have the
     bpm version followed by the number of entries. Comments are denoted by #
     
     Example of the header:
     
     # Header - libbpm version, number of BPMs
     0.1    21
     
     @param file A FILE pointer to the stream to load from
     @param version Pointer to a double that is filled with the version number
     @param num_structs Pointer to an integer that is filled with the number of 
                        structs in the file
     
     @return BPM_SUCCESS upon success, BPM_FAILURE upon failure
  */
  EXTERN int get_header( FILE *file, double *version, int *num_structs );


  /**
    Load in a structure from a file and return the arguments and the values in a list.
    Comments are denoted by #
    
    Example of a structure:
    
    # Describe x9 using a bpmconf struture
    x9
    
    @param file A FILE pointer to the stream to load from
    @param arg_list Pointer to an array of names that will hold the arguments
    @param val_list Pointer to an array of the values for each field specfied in arg_list
    @param num_args Pointer to an integer that will hold the number of arguments found

    @return BPM_SUCCESS upon success, BPM_FAILURE upon failure
  */
  EXTERN int load_struct( FILE *file, char ***arg_list, char ***val_list, int *num_args );

  
  /**
    Save a set of waveforms
    
    @param fname The filename to save to
    @param sigs The bpmsignal structures to save
    @param num_evts The number of events
    
    @return BPM_SUCCESS upon success, BPM_FAILURE upon failure
  */
  EXTERN int save_signals( char *fname, bpmsignal_t *sigs, int num_evts );


  /**
     Load the specified number of events from the given file
     
     @param fname The filename to load from
     @param sigs The bpmsignal structures
     
     @return BPM_SUCCESS upon success, BPM_FAILURE upon failure
  */
  EXTERN int load_signals( char *fname, bpmsignal_t **sigs );
  

  

#ifdef __cplusplus
}
#endif


#endif /* #ifndef BPMINTERFACE_H__ */
/*@}*/
/* ================================ end of file ============================= */
