/**
   @file
   @ingroup processing
*/
#include <math.h>

#include <bpm/bpm_messages.h>
#include <bpm/bpm_process.h>

/**
   Find the mean pedestal using the first 20 (or how ever many are required)
   sample values
   
   @param wf      a pointer to the waveform data
   @param ns      the number of samples in the waveform
   @param range   the maximum sample to go to average over
   @param nbits   the number of digitiser bits
   @param offset  returns the mean value of the samples, so voltage offset (pedestal value)
   @param rms     returns the RMS on that
   @param max     returns max value of wf
   @param min     returns min value of wf
   @param unsat_sample returns last unsaturated sample
   
   @return BPM_SUCCESS upon success, BPM_FAILURE upon failure
*/
int basic_stats( int *wf, int ns, int range, int nbits, double *offset, double *rms , 
		 int *max, int *min, int *unsat_sample) {
  
  double avg = 0., sqr_sum = 0.;
  int i;

  if ( ! wf || ! offset ) {
    bpm_error( "Invalid pointer argument in get_pedestal(...)",
	       __FILE__, __LINE__ );
    return BPM_FAILURE;
  }
  
  if ( range == 0 || ns == 0 ) {
    bpm_error("Maximum range or number of samples equal zero in get_pedestal(...)",
              __FILE__, __LINE__ );
    return BPM_FAILURE;
  }

  if ( ns < range ) {
    bpm_warning("Number of samples < max range in get_pedestal(...)", 
                __FILE__, __LINE__ );
    range = ns;
  }
  
  // find the mean and keep track of sum of squares
  *min = 1e6;
  *max = 0;
  for ( i = 0; i < range; i++) {
    avg     += (double) wf[i];
    sqr_sum += (double) wf[i]*wf[i];
    if( wf[i] > *max) *max = wf[i];
    if( wf[i] < *min) *min = wf[i];
  }

  *offset = avg / (double) range;
  if ( rms ) {
    *rms = sqrt( sqr_sum / (double) range - SQR( *offset ) );
  } else {
    bpm_warning( "NULL pointer given for rms in get_pedestal(...)",
		 __FILE__, __LINE__ );
  }

  handle_saturation( wf, ns, ns - 20, nbits, 64, unsat_sample );

  return BPM_SUCCESS;
}
