/** 
    @file  
    @ingroup processing
    @brief   libbpm main processing routines
    
    This header contains the definitions for libbpm's main BPM processing routines
*/

/** @addtogroup processing */
/** @{ */

#ifndef BPMPROCESS_H__
#define BPMPROCESS_H__

/* -----------------------------------------------------------------------------
// includes
// -------------------------------------------------------------------------- */
#include <float.h>
#include <math.h>
#include <bpm/bpm_defs.h>
#include <bpm/bpm_interface.h>


/* -----------------------------------------------------------------------------
// macro definitions
// -------------------------------------------------------------------------- */
#define  PROC_DEFAULT         0           //! default mode
#define  PROC_DO_FFT          0x00000001  //! do fft + fit for fft_freq and fft_tdecay
#define  PROC_DO_FIT          0x00000002  //! do fit, gets automatically freq + tdecay
#define  PROC_DO_DDC          0x00000004  //! do ddc, needs to know where to get freq + tdecay
#define  PROC_DDC_CALIBFREQ   0x00000008  //! ddc gets freq from calib structure
#define  PROC_DDC_CALIBTDECAY 0x00000010  //! ddc gets freq from calib structure
#define  PROC_DDC_FITFREQ     0x00000020  //! ddc gets freq from fit
#define  PROC_DDC_FITTDECAY   0x00000040  //! ddc gets tdecay from fit
#define  PROC_DDC_FFTFREQ     0x00000080  //! ddc gets freq from fft
#define  PROC_DDC_FFTTDECAY   0x00000100  //! ddc gets freq from tdecay
#define  PROC_DDC_STOREFULL   0x00000200  //! store the full DDC waveform as well
#define  PROC_FIT_DDC         0x00000400  //! fit the ddc for tdecay

/* -----------------------------------------------------------------------------
// typedefs, enums and other declarations
// -------------------------------------------------------------------------- */
#ifdef __cplusplus
extern "C" {
#endif


/* -----------------------------------------------------------------------------
// function prototypes and declarations
// -------------------------------------------------------------------------- */

  EXTERN int process_diode( bpmconf_t*, bpmsignal_t*, bpmproc_t* );
  // determines the t0 and stores in bpmproc_t

  EXTERN int process_waveform( enum bpmtype_t type, 
			       bpmconf_t *bpm, bpmcalib_t *cal, bpmsignal_t *sig, 
			       bpmproc_t *proc, bpmproc_t *trig, unsigned int mode );
  // general amplitude and phase determination according to the mode bit pattern
  // both process_monopole and process_dipole use this routine as core processing

  EXTERN int process_monopole( bpmconf_t *bpm, bpmcalib_t *cal, bpmsignal_t *sig, bpmproc_t *proc,
			       bpmproc_t *trig, unsigned int mode );
  // nothing more than a wrapper around process_waveform since there isn't much more to be
  // done

  EXTERN int process_dipole( bpmconf_t *bpm, bpmcalib_t *cal, bpmsignal_t *sig, bpmproc_t *proc, 
			     bpmproc_t *trig, bpmproc_t *ref, unsigned int mode );
  // processes the waveform and goes on to calcualte I,Q and postions and tilt


  // fitting routines
  EXTERN int fit_waveform( int *wf, int ns, double t0, double fs,
			   double i_freq, double i_tdecay, double i_amp, double i_phase,
			   double *freq, double *tdecay, double *amp, double *phase );


  EXTERN int fit_diodepulse( int *wf, int ns, double fs, double *t0 );
  // made obsolete by get_t0 ? ... yess ! implemented this routine as a wrapper around
  // get_t0 for name - consistency


  EXTERN int fit_ddc( double *ddc, int ns, double *tdecay );
  // fits the ddc to get the decay time, gets inital pars from ddc wf itself


  EXTERN int fit_fft_prepare( double **fft, int ns, double fs, int *n1, int *n2,
			      double *amp, double *freq, double *fwhm );
  // gets some initial parameters for the FFT fit...


  EXTERN int fit_fft( double **fft, int ns, double fs, double *freq, double *tdecay,
		      double *A, double *C );
  // fit the fft to get the frequency and the decay time



  // FFT routines
  EXTERN int fft_waveform( int *wf, int ns, double **fft );
  EXTERN int fft_waveform_double( double *wf, int ns, double **fft);

  // DDC routines
  EXTERN int handle_saturation( int *wf, int ns, int imax, int nbits, int threshold, 
				int *iunsat );

  EXTERN int downmix_waveform( double *wf, int ns, double fs, double freq, double t0, 
			       double **out );

  EXTERN int ddc_gaussfilter_step( double **ddc, int ns, double fs, int istart, int istop, 
				   double tfilter, double filtBW, double *out );

  EXTERN int ddc_gaussfilter( double **ddc, int ns, double fs, double filtBW, double epsFilt,
			      double **out ); 

  EXTERN int ddc_waveform( int *wf, int ns,int nbits, double fs, double t0,
			   double freq, double tdecay, double filtBW, double epsFilt,
			   double **out );
  
  EXTERN int ddc_sample_waveform( int *wf, int ns, int nbits, double fs, 
				  double t0, double t0Offset, 
				  double freq, double tdecay, double filtBW, double epsFilt,
				  double *amp, double *phase );


  // Here start the low - level routines...
  EXTERN int get_pedestal( int *wf, int ns, int range, double *offset, double *rms );

  EXTERN int basic_stats( int *wf, int ns, int range, int nbits, double *offset, double *rms , 
			  int *max, int *min, int *unsat_sample);

  EXTERN int int_to_double_waveform( double *wf_double, int *wf_int, int ns);
  EXTERN int copy_waveform( double *wf_src, double *wf_dst, int ns);
  
  EXTERN int add_scalar_waveform( double *wf, int ns, double add);
  EXTERN int mult_scalar_waveform( double *wf, int ns, double mult);
  EXTERN int mult_waveform( double *wf1, double *wf2, int ns);
  
  EXTERN int get_t0( int *wf, int ns, double fs, double *t0 );

  // some position, IQ routines...
  EXTERN int get_IQ( double amp, double phase, double refamp, double refphase,
		     double *Q, double *I );
  // gets the I and Q from the amp, phase and ref 

  EXTERN int get_pos( double Q, double I, double IQphase, double posscale,
		      double *pos );
  // returns the position given I,Q, IQphase and scale
  
  EXTERN int get_slope( double Q, double I, double IQphase, double slopescale,
			double *slope );
  // returns the slope given I,Q, IQphase and scale

  // easy conversion between time/frequency and sample number
  EXTERN int time_to_sample( double fs, int ns, double t, int *iS );
  EXTERN int sample_to_time( double fs, int ns, int iS, double *t );
  EXTERN int freq_to_sample( double fs, int ns, double f, int *iS );
  EXTERN int sample_to_freq( double fs, int ns, int iS, double *f );


#ifdef __cplusplus
}
#endif

#endif /* #ifndef BPMINTERFACE_H__ */
/*@}*/
/* ================================ end of file ============================= */
