/**
   @file
   @ingroup processing
*/

#include <bpm/bpm_messages.h>
#include <bpm/bpm_process.h>


/**
   Applies a gaussian filter to the total waveform with the given filter bandwidth
   and cut-off parameters
   
   @param ddc  complex double array with the downconverted waveform
   @param ns   number of samples
   @param fs   sampling frequency
   @param filtBW  filter bandwidth in MHz
   @param epsFilt filter cutoff parameter
   @param out    complex double array with the filtered waveform
   
   @return BPM_SUCCESS upon succcess, BPM_FAILURE upon failure
*/
int ddc_gaussfilter( double **ddc, int ns, double fs, double filtBW, double epsFilt,
		     double **out ) {

  double omega, dt, tS, tstart, tstop, filt[2];
  int istart, istop, iS;

  if ( ! ddc || ! out ) {
    bpm_error( "Invalid pointer arguments for ddc_gaussfilter(...)",
	       __FILE__, __LINE__ );
    return BPM_FAILURE;
  }

  // get frequency in rad/s
  omega = 2.*PI*filtBW;

  // determine cut-off time from cut-off parameter epsFilt
  dt = sqrt( 2.*log( omega / ( sqrt(2.*PI) * epsFilt ) ) ) / omega;

  for ( iS = 0; iS<ns; iS++ ) {
    // get time of sample we're currently doing
    sample_to_time( fs, ns, iS, &tS ); 

    // filter start and stop times
    tstart  = tS - dt;
    tstop   = tS + dt;
    
    // convert them to the sample number
    time_to_sample( fs, ns, tstart, &istart );
    time_to_sample( fs, ns, tstop, &istop );
    
    filt[Re] = 0.;
    filt[Im] = 0.;

    // perform the step
    if ( ddc_gaussfilter_step( ddc, ns, fs, istart, istop, tS, filtBW, filt ) == BPM_FAILURE ) {
      bpm_error( "Error executing step in ddc_gaussfilter(...)", __FILE__, __LINE__ );
      return BPM_FAILURE;
    }

    // filling output array
    out[iS][Re] = filt[Re];
    out[iS][Im] = filt[Im];

  } /* for ( iS = 0; iS<ns; iS++ ) */

  return BPM_SUCCESS;
}
