/**
   @file
   @ingroup processing
*/
#include <stdio.h>
#include <stdlib.h>

#include <bpm/bpm_messages.h>
#include <bpm/bpm_process.h>

/*
  Performs a fast fourier transform of the waveform, after subtracting the 
  pedestal
  
  @param *wf  the waveform
  @param  ns  the number of samples
  @param  fft the complex returned fft spectrum
  
  @return BPM_SUCCESS upon success, BPM_FAILURE upon failure
*/
int fft_waveform_double( double *wf, int ns, double **fft) {

  double *cp, ped, rms;
  int i;
  
  if ( ! wf || ! fft ) {
    bpm_error( "Invalid waveform pointers in fft_waveform(...)",
	       __FILE__, __LINE__ );
    return BPM_FAILURE;
  }


  // first make copy of waveform
  cp = (double*) calloc( 2*ns+1, sizeof(double) );
  if( ! cp ) {
    bpm_error( "Could not allocate memory in fft_waveform(...)",
	       __FILE__, __LINE__ );
    return BPM_FAILURE;
  }

  for( i=0; i<ns; i++ ) { 
    // translate to what nr_four1 needs, so build up the array + 
    // sub tract the pedestal
    cp[2*i+1] = wf[i];  // real part
    cp[2*i+2] = 0.;     // imaginary part
  }

  if ( nr_four1( cp, ns, 1 ) == BPM_FAILURE ) {
    free(cp);
    bpm_error( "Unable to execute nr_four routine in fft_waveform(...)",
	       __FILE__, __LINE__ );
    return BPM_FAILURE;
  }

  // now store in fft...
  for( i=0; i<ns; i++ ) {
    //    fft[i][Re] = cp[2*i+1];
    //    fft[i][Im] = cp[2*i+2];
    fft[i][Re] = cp[2*i+1];
    fft[i][Im] = cp[2*i+2];
  }

  free( cp );

  return BPM_SUCCESS;
}


int fft_waveform( int *intwf, int ns, double **fft ) {
  int i = 0;
  // allocate double waveform
  double *wf = (double*)calloc(ns,sizeof(double));

  // copy into double array
  for(i=0;i<ns;i++) {
    wf[i] = (double)intwf[i];
  }

  // perform fft
  fft_waveform(wf,ns,fft);

  // free double waveform
  free(wf);
}
