/**
   @file
   @ingroup processing
*/
#include <math.h>
#include <limits.h>

#include <bpm/bpm_messages.h>
#include <bpm/bpm_process.h>

/**
   Handles the saturation, so computes the first sample where no saturation
   occurs, or imax if bigger...

   @param wf the waveform
   @param ns number of samples
   @param imax maximum sample to look after
   @param nbits number of digitiser bits
   @param threshold is the distance from 0 that an adc value needs to be for it not
          to be saturated, as well as distance from 2^nbits
   @param iunsat the returned last unsaturated sample
   
   @return BPM_SUCCESS upon success, BPM_FAILURE upon failure
*/
int handle_saturation( int *wf, int ns, int imax, int nbits, int threshold, int *iunsat ) {

  int i;
  double min_value, max_value;

  *iunsat = -INT_MAX;

  if ( ! wf ) {
    bpm_error( "Invalid waveform pointer in handle_saturation(...)",
	       __FILE__, __LINE__ );
    return BPM_FAILURE;
  }

  if ( imax >= ns ) {
    bpm_warning( "Maximum sample >= number of samples in handle_saturation(...)",
		 __FILE__, __LINE__ );
    imax = ns - 1;
  }
  
  min_value = threshold;
  max_value = (int) pow( 2., nbits ) - threshold;

  if ( max_value <= min_value ) {
    bpm_error( "Check number of bits in ADC and threshold for handle_saturation(...)",
	       __FILE__, __LINE__ );
    return BPM_FAILURE;
  }

  // search for sample
  for( i = imax; i > 0; --i ) {
    if ( ( wf[i] > max_value ) || ( wf[i] < min_value ) ) break;
  }

  // check 
  if ( i == ns ) {
    bpm_warning( "Last unsaturated sample equal to number of samples in handle_saturation(...)",
		 __FILE__, __LINE__ );
    i--;
  }

  *iunsat = i + 1;

  return BPM_SUCCESS;
}
