/*
  libbpm - BPM signal processing/simulation library
  Copyright (C) 2006-07 Bino Maiheu (bino@hep.ucl.ac.uk)
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA 
*/

/**
   @file
   @ingroup sim
   @brief   libbpm waveform simulation routines

   This header contains the definitions for the libbpm RF waveform
   simulation routines
*/

/** @addtogroup sim */
/** @{ */

#ifndef BPMSIM_H__
#define BPMSIM_H__

/* -----------------------------------------------------------------------------
// includes
// -------------------------------------------------------------------------- */
#include <math.h>
#include <bpm/bpm_defs.h>
#include <bpm/bpm_interface.h>
#include <bpm/bpm_wf.h>
#include <bpm/bpm_nr.h>
#include <bpm/bpm_dsp.h>

/* -----------------------------------------------------------------------------
// macro definitions
// -------------------------------------------------------------------------- */

/* -----------------------------------------------------------------------------
// typedefs, enums and other declarations
// -------------------------------------------------------------------------- */
#ifdef __cplusplus
extern "C" {
#endif
/* -----------------------------------------------------------------------------
// function prototypes and declarations
// -------------------------------------------------------------------------- */


  // bino's new routines for absim...
  EXTERN int generate_bpmsignal( bpmconf_t *bpm, beamconf_t *beam, doublewf_t *RF );

  EXTERN int add_mode_response( complexwf_t *RF, bpmconf_t *bpm, bpmmode_t *mode, 
                                beamconf_t *beam );

  EXTERN complex_t get_mode_amplitude( bpmconf_t *bpm, bpmmode_t *mode, 
				       beamconf_t *beam );






  //EXTERN int generate_monopole( bpmconf_t*, beamconf_t*, bpmsignal_t* );

  //EXTERN int generate_dipole( bpmconf_t*, beamconf_t*, bpmsignal_t* );

  //EXTERN int generate_diode( bpmconf_t*, beamconf_t*, bpmsignal_t* );

  /**
     Calculate the response of a dipole signal given an incoming bunch
     @param bunchcharge The charge of the bunch (in nC)
     @param bunchlength The length of the bunch (in mm)
     @param pos The position of the beam (in mm)
     @param possens The position sensitivty of the BPM (in V/nC/mm)
     @param slope The slope of the beam (in rad)
     @param slopesens The slope sensitivity of the BPM (in V/nC/urad)
     @param tilt The tilt of the bunch (in urad)
     @param tiltsens The tilt sensitivty of the BPM (V/nC/urad)
     @param Amp the complex amplitude of the waveform at the arrival time
     @return BPM_SUCCES upon succes, BPM_FAILURE upon failure
  */
  EXTERN int get_dipole_amp( double bunchcharge, double bunchlength,
                             double pos, double possens,
                             double slope, double slopesens,
                             double tilt, double tiltsens,
                             complex_t *Amp );

  /**
     Calculate the response of a dipole signal given an incoming bunch   
     @param bunchcharge The charge of the bunch (in nC)
     @param bunchlength The length of the bunch (in mm)
     @param chargesens The charge sensitivty of the BPM (V/nC)
     @param Amp the complex amplitude of the waveform at the arrival time
     @return BPM_SUCCES upon succes, BPM_FAILURE upon failure
  */
  EXTERN int get_monopole_amp( double bunchcharge, double bunchlength,
                             double chargesens,
                             complex_t *Amp );

  /**
     Generates a one sample impulse to excite the resonator
     @param ttrig the trigger time
     @param RF waveform containing the impulse
     @return BPM_SUCCES upon succes, BPM_FAILURE upon failure
  */
  EXTERN int add_excitation( double ttrig, doublewf_t *RF );

  /**
     Calculates the normalized complex mode response
     @param excitation array containing the excitation profile
     @param freq mode resonant frequency
     @param Qvalue mode quality factor
     @param response complex mode response normalised to amp(t0) = 1
     @return BPM_SUCCESS upon success or BPM_FAILURE upon failure
  */
  EXTERN int get_mode_response( doublewf_t *excitation,
                       double freq, double Qvalue,
                       complexwf_t *response );

  /**
    Adds a template waveform to the signal with a given scale and rotation
    @param RF the signal
    @param TEMP template waveform to add
    @param f complex number encoding the amplitude and phase
    @return BPM_SUCCESS upon success, BPM_ERROR upon error
  */
  EXTERN int add_waveforms( complexwf_t *RF, complexwf_t *TEMP, complex_t f );

  /**
     Add the given amount of amplitude noise to a complex array
     @param amplnoise The amplitude noise to add to the waveform (in Volts)
     @param IF Complex waveform containing the signal
     @return BPM_SUCCESS upon succes, BPM_FAILURE upon failure
  */ 
  EXTERN int add_amplnoise( double amplnoise, complexwf_t *IF );

/**
   Digitises the waveform using the sampling frequency
   and the number of samples set in the resulting waveform
   @param IF input waveform to digitse
   @param nbits bit resolution of the ADC
   @param range_min the minimum voltage and
   @param range_max the maximum voltage the ADC can process
   @param clock_jitter ADC clock jitter
   @param digi_noise rms digitiser noise in ADC channels
   @param ipmode interpolation mode for doublewf_getvalue()
   @param wf sampled waveform
   @return BPM_SUCCESS upon success, BPM_FAILURE upon failure
*/
  EXTERN int digitise( doublewf_t *IF, int nbits, double range_min, double range_max,
		       double clock_jitter, double digi_noise, unsigned int ipmode, 
		       intwf_t *wf );

#ifdef __cplusplus
}
#endif

#endif /* #ifndef BPMINTERFACE_H__ */
/** @} */
/* ================================ end of file ============================= */
