/**
   @file
   @ingroup sim
*/
#include <bpm/bpm_messages.h>
#include <bpm/bpm_simulation.h>
#include <bpm/bpm_wf.h>
#include <bpm/bpm_dsp.h>

/**
   Calculate the normalized complex dipole response
   @param excitation array containing the excitation profile
   @param freq mode resonant frequency
   @param Qvalue mode quality factor
   @param response complex mode response normalised to amp(t0) = 1
   @return BPM_SUCCESS upon success or BPM_FAILURE upon failure
*/
int get_mode_response( doublewf_t *excitation,
		       double freq, double Qvalue, 
		       complexwf_t *response ) {

  if ( ! excitation || ! response ) {
    bpm_error( "Invalid pointer argument in add_wave(...)",
               __FILE__, __LINE__ );
    return BPM_FAILURE;
  }

  // mode resonator
  filter_t *resonator =
           create_filter( "resonator", RESONATOR | BANDPASS, 0,
                          excitation->ns, excitation->fs,
                          freq, 0., Qvalue );

  doublewf_t *resp_re = doublewf_copy_new( excitation );

  apply_filter( resonator, resp_re->wf );

  doublewf_scale( resp_re->fs/(2*PI*freq/Qvalue), resp_re );

  doublewf_t *resp_im = doublewf_copy_new( resp_re );

  doublewf_integrate( resp_re );
  doublewf_scale( 2*PI*freq, resp_re );

  complexwf_setreal( response, resp_re );
  complexwf_setimag( response, resp_im );

  delete_filter( resonator );
  doublewf_delete( resp_re );
  doublewf_delete( resp_im );

  return BPM_SUCCESS;
}

// end of file
